#!/usr/bin/python
"""
(C) Copyright 2020
Scott Wiederhold, s.e.wiederhold@gmail.com
https://community.openglow.org

SPDX-License-Identifier:    MIT
"""
import logging

from gfutilities.configuration import *
from gfutilities import GFUIService, Emulator

parse('gf-machine-emulator.cfg')

logging.basicConfig(format='(%(levelname)s) %(module)s:%(funcName)s %(message)s')
logger = logging.getLogger("openglow")
logger.setLevel(log_level(get_cfg('LOGGING.CONSOLE_LEVEL')))

if get_cfg('LOGGING.FILE'):
    fh = logging.FileHandler(get_cfg('LOGGING.FILE'))
    fh.setLevel(log_level(get_cfg('LOGGING.LEVEL')))
    fh.setFormatter(logging.Formatter('%(asctime)s (%(levelname)s) %(module)s:%(funcName)s %(message)s'))
    logger.addHandler(fh)

service = GFUIService(Emulator())
service.connect()
service.run()



"""
(C) Copyright 2020
Scott Wiederhold, s.e.wiederhold@gmail.com
https://community.openglow.org

SPDX-License-Identifier:    MIT
"""
from gfutilities.puls.pulsedata import decode_all_steps, generate_linear_puls

__all__ = ['decode_all_steps', 'generate_linear_puls']



"""
(C) Copyright 2020
Scott Wiederhold, s.e.wiederhold@gmail.com
https://community.openglow.org

SPDX-License-Identifier:    MIT
"""
_decode_step_codes = {
    'LE': {'mask': 0b00010000, 'test': 0b00010000},  # Laser Enable (ON)
    'LP': {'mask': 0b10000000, 'test': 0b01111111},  # Laser Power Setting
    'XP': {'mask': 0b00000011, 'test': 0b00000001},  # X+ Step
    'XN': {'mask': 0b00000011, 'test': 0b00000011},  # X- Step
    'YN': {'mask': 0b00001100, 'test': 0b00000100},  # Y- Step
    'YP': {'mask': 0b00001100, 'test': 0b00001100},  # Y+ Step
    'ZP': {'mask': 0b01100000, 'test': 0b00100000},  # Z+ Step
    'ZN': {'mask': 0b01100000, 'test': 0b01100000},  # Z- Step
}

_SPEED = 1000


def decode_all_steps(puls: bytes, data: dict = None, mode: tuple = (8, 2)) -> dict:
    cnt = {
        'XP': 0,
        'XN': 0,
        'XTOT': 0,
        'XEND': 0,
        'YP': 0,
        'YN': 0,
        'YTOT': 0,
        'YEND': 0,
        'ZP': 0,
        'ZN': 0,
        'ZTOT': 0,
        'ZEND': 0,
        'LE': 0,
        'LP': 0,
    }
    for step in puls:
        if step & _decode_step_codes['LP']['mask']:
            # Power Setting (LP)
            cnt['LP'] = cnt['LP'] + 1
        else:
            for action in sorted(_decode_step_codes):
                if not (step & _decode_step_codes[action]['mask']) ^ _decode_step_codes[action]['test']:
                    cnt[action] = cnt[action] + 1
                    if action != 'LE':
                        cnt[action[0:1] + 'TOT'] = cnt[action[0:1] + 'TOT'] + 1
                        cnt[action[0:1] + 'END'] = cnt[action[0:1] + 'END'] + 1 \
                            if action[1:2] == 'P' else cnt[action[0:1] + 'END'] - 1

    for axis in ('X', 'Y'):
        cnt[axis + 'MM'] = (cnt[axis + 'END'] / mode[0]) * 0.15
    cnt['ZMM'] = (cnt['ZEND'] / mode[1]) * 0.70612

    for axis in ('X', 'Y', 'Z'):
        cnt[axis + 'IN'] = cnt[axis + 'MM'] / 25.4

    if data is not None:
        for key, val in cnt.items():
            cnt[key] = val + data.get(key, 0)
    return cnt


def generate_linear_puls(x: int, y: int, outfile: str) -> None:
    expected_count = abs(x) if abs(x) >= abs(y) else abs(y)

    max_speed = 5
    min_speed = 55
    acc = 10
    acc_dist = int((min_speed - max_speed) / acc)
    acc_dist = acc_dist if acc_dist < (expected_count / 2) else int(expected_count / 2)

    steps = 0
    d = min_speed

    with open(outfile, 'bw') as f:
        for xs, ys in _step_gen(x, y):
            steps += 1
            s = 0
            if xs != 0:
                s |= 0b00000001 if xs > 0 else 0b00000011
            if ys != 0:
                s |= 0b00001100 if ys > 0 else 0b00000100
            f.write(bytes([s]))

            if steps <= acc_dist:
                # Accelerating
                d = d - acc if d > max_speed - acc else max_speed
            elif steps >= (expected_count - acc_dist):
                # Decelerating
                d = d + acc if d < min_speed + acc else min_speed
            f.write('\0'.encode() * d)


def _step_gen(x: int, y: int) -> tuple:
    xd = 1 if x >= 0 else -1
    yd = 1 if y >= 0 else -1
    xt = abs(x)
    yt = abs(y)

    if xt >= yt:
        maj_target = xt
        min_target = yt
        maj_dir = xd
        min_dir = yd
    else:
        maj_target = yt
        min_target = xt
        maj_dir = yd
        min_dir = xd
    if min_target > 0:
        maj_per_min = round(maj_target / min_target, 4)
    else:
        maj_per_min = maj_target

    maj_cnt = 0
    min_cnt = 0
    while maj_cnt < maj_target or min_cnt < min_target:
        if maj_cnt < maj_target:
            maj_cnt += 1
            if maj_cnt % maj_per_min < 1 and min_cnt < min_target:
                min_out = min_dir
                min_cnt += 1
            else:
                min_out = 0
        else:
            maj_dir = 0
            min_out = min_dir
            min_cnt += 1

        yield maj_dir if xt >= yt else min_out, min_out if xt >= yt else maj_dir



"""
(C) Copyright 2020
Scott Wiederhold, s.e.wiederhold@gmail.com
https://community.openglow.org

SPDX-License-Identifier:    MIT
"""
import json
import logging
import time

from gfutilities._common import *
from gfutilities.configuration import get_cfg
from gfutilities.device.basemachine import BaseMachine
from gfutilities.service.websocket import load_motion, img_upload, send_wss_event

logger = logging.getLogger(LOGGER_NAME)


class Emulator(BaseMachine):
    """
    Glowforge Device Emulator
    This connects to the servers specified in the configuration, and responds the same as the hardware does.
    """
    def __init__(self):
        """
        Initialize object
        """
        BaseMachine.__init__(self)
        # Track which step in the homing process we are at.
        # Example script has 5 saved head images, which are uploaded to the service in sequence to
        # emulate the homing process.  Files are located in _RESOURCES/IMG
        # HOME_1.jpg: Image of the head located at expected center position
        #             This image was taken with the head already in the calibrated position to speed up the process.
        # HOME_2.jpg: Image after first retreat move
        # HOME_3.jpg: Image after second retreat move
        # HOME_4.jpg: Image after third retreat move
        # LID_IMAGE.jpg: Image of bed with Proofgrade(tm) material
        #                This image is sent for every none-homing lid_image request
        self._homing_stage = 0 if get_cfg('EMULATOR.BYPASS_HOMING') else 1
        # Used to track the action just before a lid_image request. This is used to detect homing operations after
        # the initial start up (after a print job, for instance)
        self._last_action = None

    def _button_wait(self, msg: dict) -> None:
        pass

    def _head_image(self, msg: dict, settings: dict = None) -> None:
        if settings is not None and settings.get('HCil') > 0:
            img = 'HEAD_LASER_%s.jpg' % get_cfg('EMULATOR.MATERIAL_THICKNESS')
        else:
            img = 'HEAD_NO_LASER_%s.jpg' % get_cfg('EMULATOR.MATERIAL_THICKNESS')
        with open('%s/%s' % (get_cfg('EMULATOR.IMAGE_SRC_DIR'), img), 'rb') as f:
            img_upload(self._session, f.read(), msg)
        self._last_action = 'head_image'

    def _hunt(self, msg: dict) -> None:
        self._last_action = 'hunt'
        self._motion_download(msg)

    def _initialize(self) -> None:
        pass

    def _lid_image(self, msg: dict) -> None:
        if self._homing_stage == 0 and self._last_action == 'motion':
            # Looks like were are homing again
            self._homing_stage = 1
        if self._homing_stage != 0:
            logger.info('HOME Step %s' % self._homing_stage)
            img = 'HOME_%s.jpg' % self._homing_stage
            self._homing_stage = self._homing_stage + 1 if self._homing_stage < 4 else 0
        else:
            img = 'LID_IMAGE.jpg'
        with open('%s/%s' % (get_cfg('EMULATOR.IMAGE_SRC_DIR'), img), 'rb') as f:
            img_upload(self._session, f.read(), msg)
        self._last_action = 'lid_image'

    def _motion(self, msg: dict) -> None:
        # Download puls file from service
        self._motion_download(msg)
        if msg['action_type'] == 'print':
            send_wss_event(self._q_msg_tx, msg['id'], 'print:download:completed')
            self._button_wait(msg)
            if not self._running_action_cancelled:
                send_wss_event(self._q_msg_tx, msg['id'], 'print:warmup:starting')
        # Run motion job
        if not self._running_action_cancelled:
            if msg['action_type'] == 'print':
                send_wss_event(self._q_msg_tx, msg['id'], 'print:running')
                send_wss_event(self._q_msg_tx, msg['id'], 'print:return_to_home:succeeded')

        self._last_action = msg['action_type']

    def _motion_download(self, msg: dict):
        logger.info('DOWNLOADING')
        base_file_name = '%s/%s_%s' % (get_cfg('EMULATOR.MOTION_DL_DIR'),
                                       time.strftime("%Y-%m-%d_%H%M%S"), msg['action_type'])
        info = load_motion(self._session, msg['motion_url'], base_file_name + '.puls')
        with open(base_file_name + '.info', 'w') as f:
            f.write(json.dumps(info, sort_keys=True, indent=4, default=str) + '\n')
        logger.debug('Header: %s' % info)
        logger.info('COMPLETE')
        return

    def _shutdown(self) -> None:
        pass



"""
(C) Copyright 2020
Scott Wiederhold, s.e.wiederhold@gmail.com
https://community.openglow.org

SPDX-License-Identifier:    MIT
"""
from datetime import timedelta
import json
import logging
from pathlib import Path
from queue import Queue
from requests import Request, Response, Session
from threading import Thread
import time
from typing import Union

from lomond import WebSocket
from lomond.persist import persist

from gfutilities._common import *
from gfutilities.configuration import get_cfg, set_cfg
from gfutilities.puls import decode_all_steps

start_time = time.time()
response_id = 31

logger = logging.getLogger(LOGGER_NAME)


class WsClient(Thread):
    """
    Web Socket Client
    Establishes a persistent WSS client that launches a separate thread to handle WSS operations.
    Spools incoming messages to the q_msg_rx Queue, and sends any messages placed into the q_msg_tx Queue.
    """
    def __init__(self, q_rx: Queue, q_tx: Queue):
        """
        Class Initializer
        :param q_rx: WSS RX Message Queue
        :type q_rx: Queue
        :param q_tx: WSS TX Message Queue
        :type q_tx: Queue
        """
        self.msg_q_rx = q_rx
        self.msg_q_tx = q_tx
        self.stop = False
        self.ready = False
        self.ws = WebSocket(url=get_cfg('SERVICE.STATUS_SERVICE_URL') + '/' + get_cfg('SESSION.WS_TOKEN'),
                            protocols=['glowforge'], agent=get_cfg('SESSION.USER_AGENT'))
        self.ignore_events = ['ping', 'pong', 'poll', 'connecting', 'connected']
        Thread.__init__(self)

    def run(self) -> None:
        """
        Thread loop
        Listens for WS messages, and adds them the q_msg_rx Queue.
        Monitors the q_msg_tx, and sends any messages it finds.
        Handles PING, PONG, and POLL messages.
        :return:
        """
        for event in persist(self.ws):
            if event.name not in self.ignore_events:
                logger.info('RX-EVENT: ' + event.name)
                logger.debug(event)
            if self.stop:
                logger.info('STOP REQUESTED')
                break
            # WS reporting session is established
            if event.name == 'ready':
                self.ready = True
            # Service has sent us a text message
            elif event.name == 'text':
                logger.debug(event.text)
                self.msg_q_rx.put(event.text)
            elif event.name not in self.ignore_events:
                logger.error('UNKNOWN RX-EVENT: ' + event.name)
            while self.ready:
                if not self.msg_q_tx.empty():
                    send_msg = self.msg_q_tx.get()
                    logger.info('TX-EVENT: ' + send_msg.strip())
                    self.ws.send_text(send_msg)
                    self.msg_q_tx.task_done()
                else:
                    break
        logger.info('CLOSING')
        self.ws.close()


def _byte_to_int(data: bytes) -> int:
    """
    Returns integer value of big endian byte data
    :param data:
    :return: Integer value
    :rtype: int
    """
    return ord(data[0:1]) + (ord(data[1:2]) * 0x100) + (ord(data[2:3]) * 0x10000) + (ord(data[3:4]) * 0x1000000)


def firmware_check(s: Session) -> Union[dict, bool]:
    """
    Checks for the latest version of available firmware.
    Returns a {dict} if new firmware is available.
    {'version': '<new version>', 'download_url': '<url to download firmware>'}
    :param s: Requests Session object
    :type s: Session
    :return: Returns False if configured firmware is equal latest.
    :rtype: Union[dict, bool]
    """
    logger.info('CHECKING')
    r = request(s, get_cfg('SERVICE.SERVER_URL') + '/update/current', 'GET')
    if not r:
        logger.error('FAILED')
        return False
    rj = r.json()
    logger.debug(rj)
    if rj['version'] is not None and not rj['version'] == get_cfg('FACTORY_FIRMWARE.FW_VERSION'):
        logger.info('New Firmware Available: Installed (%s), Available (%s)'
                    % (get_cfg('FACTORY_FIRMWARE.FW_VERSION'), rj['version']))
        return rj
    else:
        logger.info('LATEST ALREADY INSTALLED.')
        return False


def firmware_download(s: Session, fw: dict) -> bool:
    """
    Downloads latest firmware to directory set in config.
    :param s: Requests Session object
    :type s: Session
    :param fw: Dictionary returned by _api_firmware_check()
    :type fw: dict
    :return: {bool} True
    :rtype: bool
    """
    logger.info('DOWNLOADING')
    Path(get_cfg('FACTORY_FIRMWARE.DOWNLOAD_DIR')).mkdir(parents=True, exist_ok=True)
    r = request(s, fw['download_url'], 'GET', stream=True)

    with open('%s/glowforge-fw-v%s.fw' % (get_cfg('FACTORY_FIRMWARE.DOWNLOAD_DIR'), fw['version']), 'wb') as f:
        for chunk in r.iter_content(chunk_size=1024):
            if chunk:
                f.write(chunk)
    logger.info('COMPLETE.')
    return True


def get_session() -> Session:
    """
    Creates web api session object.
    :return: Requests Session object
    :rtype: Session
    """
    s = Session()
    set_cfg('SESSION.USER_AGENT', 'OpenGlow/%s' % get_cfg('FACTORY_FIRMWARE.FW_VERSION'))
    s.headers.update({'user-agent': get_cfg('SESSION.USER_AGENT')})
    logger.debug('Returning object : %s' % s)
    return s


def img_upload(s: Session, img: bytes, msg: dict) -> bool:
    """
    Uploads head/lid image
    :param s: Requests Session object
    :type s: Session
    :param img: Path to file to upload, or bytes object containing image
    :type img: bytes
    :param msg: WSS message
    :type msg: dict
    :return:
    :rtype: bool
    """
    logger.info('START')
    url = get_cfg('SERVICE.SERVER_URL') + '/api/machines/%s/%s' % (msg['action_type'], msg['id'])
    headers = {'Content-Type': 'image/jpeg'}
    r = request(s, url, 'POST', data=img, headers=headers)
    if not r:
        logger.debug(r)
        return False
    else:
        logger.debug(r.json)
    logger.info('COMPLETE')
    return True


def load_motion(s: Session, url: str, out_file: str) -> Union[dict, bool]:
    """
    Downloads motion/print file, parses header
    :param s: Requests Session object
    :type s: Session
    :param url: Target URL
    :type url: str
    :param out_file: Where to write the results
    :type out_file: str
    :return: Header dict or False on failure
    :rtype: Union[dict, bool]
    """
    r = request(s, url, 'GET', stream=True)
    info = {
        'header_data': {},
        'header_len': 0,
        'size': 0,
        'run_time': None,
    }
    with open(out_file, 'wb') as f:
        res = r.iter_content(chunk_size=1024)
        puls_data = next(res)
        if puls_data[1:4].decode() != 'GF1':
            logger.error('received data not a GF puls file')
            return False
        info['header_len'] = _byte_to_int(puls_data[4:8]) - 8
        s = 8
        while s < (info['header_len'] + 8):
            info['header_data'][puls_data[s:s + 4].decode()] = _byte_to_int(puls_data[s + 4:s + 8])
            s += 8
        size = len(puls_data[s:])
        stat = decode_all_steps(puls_data[s:])
        # TODO: REMOVE THIS TEMP RAW WRITE - ONLY FOR QUICK TESTING
        base_file_name = '%s/%s' % (str(Path(get_cfg('LOGGING.FILE')).parent), time.strftime("%Y-%m-%d_%H%M%S"))
        raw = open(base_file_name + '.puls', 'wb')
        raw.write(puls_data)
        f.write(puls_data[s:])
        for chunk in res:
            if chunk:
                size = size + len(chunk)
                stat = decode_all_steps(chunk, stat)
                f.write(chunk)
                raw.write(chunk)
        raw.close()
        info['size'] = size
        info['run_time'] = timedelta(seconds=size / info['header_data']['STfr'])
        info['stats'] = stat
        raw = open(base_file_name + '.info', 'w')
        raw.write(json.dumps(info, sort_keys=True, indent=4, default=str) + '\n')
        raw.close()
        return info


def request(s: Session, url: str, method: str, timeout: int = 15, stream: bool = False, **kwargs) \
        -> Union[Response, bool]:
    """
    Submits requests to provided url using the established Session object
    :param s: Requests Session object
    :type s: Session
    :param url: Target URL
    :type url: str
    :param method: Method for the request
    :type method: str
    :param timeout: Timeout
    :type timeout: int
    :param stream: Stream
    :type stream: bool
    :param kwargs:
    :return: Requests Response object, or False on error.
    :rtype: Union[Response, bool]
    """
    req = Request(method, str.replace(str(url), 'wss://', 'https://'), **kwargs)
    req = s.prepare_request(req)
    r = s.send(req, stream=stream, timeout=timeout)
    if r.status_code != 200:
        logger.error('FAILED: ' + r.reason)
        return False
    else:
        return r


def send_wss_event(msg_q_tx: Queue, action_id: Union[int, None], event: str, **kwargs):
    global response_id
    global start_time
    response_id += 1

    if 'data' in kwargs:
        data = ',"%s":%s' % (kwargs['data']['key'], kwargs['data']['value'])
    else:
        data = ''
    if action_id is None:
        action_id = ''
    else:
        action_id = '"action_id":%s,' % action_id
    msg_q_tx.put(u'{"id":%s,"timestamp":%s,"type":"event","version":1,'
                 u'%s"level":"INFO","event":"%s"%s}\n' %
                 (response_id, int(((time.time() - start_time) * 100) + 3800), action_id, event, data))


def update_header(s: Session, header: str, value: str) -> bool:
    """
    Adds persistent header to Session
    :param s: Requests Session object
    :type s: Session
    :param header: Header Name
    :type header: str
    :param value: Header Value
    :type value: str
    :return: True
    :rtype: bool
    """
    s.headers.update({header: value})
    return True


def ws_connect(msg_q_rx: Queue, msg_q_tx: Queue) -> bool:
    """
    Establishes Web Socket Session
    :param msg_q_rx: WSS RX essage Queue
    :type msg_q_rx: Queue
    :param msg_q_tx: WSS TX Message Queue
    :type msg_q_tx: Queue
    :return: Web Socket session object
    :rtype: WebSocket
    """
    logger.info('CONNECTING')
    ws = WsClient(msg_q_rx, msg_q_tx)
    ws.start()
    ws_ready = False
    # Wait 15 seconds for session to establish, or error out
    for x in range(0, 16):
        if not ws.ready:
            time.sleep(1)
        else:
            ws_ready = True
            break
    if ws_ready:
        logger.info('ESTABLISHED')
        return True
    else:
        logger.error('FAILED')
        ws.stop = True
        return False



"""
(C) Copyright 2020
Scott Wiederhold, s.e.wiederhold@gmail.com
https://community.openglow.org

SPDX-License-Identifier:    MIT
"""
from gfutilities.device.basemachine import BaseMachine
from gfutilities.device.emulator import Emulator
from gfutilities.service.gfuiservice import GFUIService

__all__ = ['GFUIService', 'Emulator', 'BaseMachine']



"""
(C) Copyright 2020
Scott Wiederhold, s.e.wiederhold@gmail.com
https://community.openglow.org

SPDX-License-Identifier:    MIT
"""
import logging
from requests import Session
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from gfutilities._common import *
from gfutilities.configuration import get_cfg, set_cfg
from gfutilities.service.websocket import request, update_header

logger = logging.getLogger(LOGGER_NAME)


def authenticate_machine(s: Session) -> bool:
    """
    Authenticates machine to service.
    POSTS serial number prefixed with an S and machine password to <server_url>/machines/sign_in.
    On success, server responds with JSON encoded 'ws_token' for WSS auth, and 'auth_token' that is
    added as a header to all session web API requests.
    Auth token header is 'Authorization': 'Bearer <auth_token>'
    :param s: Emulator session object.
    :type s: Session
    :return: Authentication result.
    :rtype: bool
    """
    logger.info('START')

    Retry.BACKOFF_MAX = 30
    retries = Retry(
        total=30,
        backoff_factor=5,
        method_whitelist=["GET"]
    )
    adapter = HTTPAdapter(max_retries=retries)
    s.mount("https://", adapter)

    r = request(s, get_cfg('SERVICE.SERVER_URL') + '/machines/sign_in', 'POST',
                data={'serial': 'S' + str(get_cfg('MACHINE.SERIAL')), 'password': get_cfg('MACHINE.PASSWORD')})
    if r:
        rj = r.json()
        logger.debug(rj)
        set_cfg('SESSION.WS_TOKEN', rj['ws_token'])
        set_cfg('SESSION.AUTH_TOKEN', rj['auth_token'])
        update_header(s, 'Authorization', "Bearer %s" % get_cfg('SESSION.AUTH_TOKEN'))
        logger.info('SUCCESS')
        return True
    else:
        logger.error('FAILED')
        return False



"""
(C) Copyright 2020
Scott Wiederhold, s.e.wiederhold@gmail.com
https://community.openglow.org

SPDX-License-Identifier:    MIT
"""
import configparser
import logging
from typing import Any

_CONFIG = {}


def parse(cfg_file: str) -> None:
    """
    Parses configuration file and returns dict object with values as: <SECTION>.<PARAM>: value.
    Note: SECTION and PARAM are converted to uppercase.
    :param cfg_file: path to configuration file
    :type cfg_file: str
    """
    defaults = {
        'log_level': 'DEBUG',
        'console_log_level': 'False',
    }
    config = configparser.ConfigParser(defaults)
    config.read(cfg_file)
    for section in config.sections():
        for item in config.items(section):
            if item[1] == 'True':
                value = True
            elif item[1] == 'False':
                value = False
            else:
                value = item[1]
            _CONFIG['%s.%s' % (str.upper(section), str.upper(item[0]))] = value


def log_level(level_str: str) -> int:
    """
    Returns logging log level object based on provided string.
    :param level_str: Desired logging level
    :type level_str: str
    :return: Logging level object
    :rtype: int
    """
    levels = {
        'CRITICAL': logging.CRITICAL, 'ERROR': logging.ERROR, 'WARNING': logging.WARNING, 'INFO': logging.INFO,
        'DEBUG': logging.DEBUG
    }
    if level_str in levels:
        return levels[level_str]
    else:
        return logging.INFO


def get_cfg(parameter: str) -> Any:
    """
    Returns value of configuration parameter
    :param parameter: Parameter to return
    :type parameter: str
    :return: The value of the parameter
    :rtype: Any
    """
    return _CONFIG.get(parameter)


def set_cfg(parameter: str, value: Any, keep_value: bool = False):
    """
    Returns value of configuration parameter
    :param parameter: Parameter to set
    :type parameter: str
    :param value: Value of parameter
    :type value: Any
    :param keep_value: Keep existing value if set
    :type keep_value: bool
    """
    if keep_value and get_cfg(parameter) is not None:
        return
    _CONFIG[parameter] = value


__all__ = ['get_cfg', 'parse', 'set_cfg', 'log_level']



"""
(C) Copyright 2020
Scott Wiederhold, s.e.wiederhold@gmail.com
https://community.openglow.org

SPDX-License-Identifier:    MIT
"""
import logging
from queue import Queue
from requests import Session
from threading import Thread
from typing import Union

from gfutilities._common import *
from gfutilities.configuration import set_cfg
from gfutilities.device.settings import send_report, get_machine_setting
from gfutilities.service.websocket import send_wss_event

logger = logging.getLogger(LOGGER_NAME)


class BaseMachine:
    """
    BaseMachine Base Class
    Serves as the foundation for emulator and machine control classes.
    Implements the basic functions required for operation with the Glowforge online service.
    """
    def __init__(self):
        """
        Initializes the object, child threads and message queues.
        """
        self._action_thread: _ActionThread = _ActionThread(self, {})
        self._q_msg_tx: Union[Queue, None] = None
        self.running_action_id: Union[int, None] = 0
        self.running_action_type: Union[str, None] = None
        self._running_action_cancelled: bool = False
        self._session: Session = Union[Queue, None]

        set_cfg('FACTORY_FIRMWARE.FW_VERSION', get_machine_setting('MCov'), True)
        set_cfg('FACTORY_FIRMWARE.APP_VERSION', get_machine_setting('MCdv'), True)

    def _ok_to_run_action(self, action_id: str, msg_type: str, status: str) -> bool:
        """
        Checks the action against any current running actions.
        If this a request to cancel an action:
            - If the action is currently running, it sets the _running_action_cancelled flag as True
            - If the action is not currently running, it sends a cancelled message to the service.
            - For both conditions, it returns False
        If there are not any currently running actions, it sets this as the current running action and returns True.
        If there is a current running action, and this is not a request to cancel it, sends a cancelled message
        to the service for this requested action, and returns false.
        :param action_id: Action ID
        :type action_id: str
        :param msg_type: Action Type
        :type msg_type: str
        :param status: Status of message (either 'ready' or 'cancelled')
        :type status: str
        :return: Return True if this it is ok to run this action, False if not
        :rtype: bool
        """
        action_id = int(action_id)
        if action_id == self.running_action_id and status == 'cancelled':
            logger.debug('action %s (%s) cancellation received' % (action_id, msg_type))
            self._running_action_cancelled = True
            return False
        elif self.running_action_id or status == 'cancelled':
            return False
        else:
            logger.debug('running action set to %s (%s)' % (action_id, msg_type))
            self.running_action_id = action_id
            self.running_action_type = msg_type
            self._running_action_cancelled = False
            return True

    def _button_wait(self, msg: dict) -> None:
        """
        Child class handler for "print:waiting" event - waiting for button push
        To be implemented by the child class.
        This is method should return after the big button has been pressed.
        :param msg: Incoming WSS Message
        :type msg: dict
        :return:
        """
        raise NotImplementedError

    def head_image(self, msg: dict) -> None:
        """
        Process head image request.
        Sends related start and finish messages, and calls child class handler.
        :param msg: Incoming WSS Message
        :type msg: dict
        :return:
        """
        send_wss_event(self._q_msg_tx, msg['id'], 'head_image:starting')
        self._head_image(msg)
        send_wss_event(self._q_msg_tx, msg['id'], 'head_image:completed')

    def _head_image(self, msg: dict, settings: dict = None) -> None:
        """
        Child class handler for capturing image from head cam.
        To be implemented by the child class.
        This method should capture the image from the head camera, and upload the resulting image to the Web API.
        :param msg: Incoming WSS Message
        :type msg: dict
        :param settings: Camera settings
        :type settings: dict
        :return:
        """
        raise NotImplementedError

    def hunt(self, msg: dict) -> None:
        """
        Home the focus lens.
        Sends related start and finish messages, and calls child class handler.
        :param msg: Incoming WSS Message
        :type msg: dict
        :return:
        """
        send_wss_event(self._q_msg_tx, msg['id'], 'hunt:starting')
        self._hunt(msg)
        send_wss_event(self._q_msg_tx, msg['id'], 'hunt:completed')

    def _hunt(self, msg: dict) -> None:
        """
        Child class handler for focus lens homing cycle
        To be implemented by the child class.
        This method should home the lens and set it to the appropriate zeroing offset.
        :param msg: Incoming WSS Message
        :type msg: dict
        :return:
        """
        raise NotImplementedError

    def _initialize(self) -> None:
        """
        Child class handler for initializing the machine
        To be implemented by the child class.
        :return:
        """
        raise NotImplementedError

    def lid_image(self, msg: dict) -> None:
        """
        Process lid image request.
        Sends related start and finish messages, and calls child class handler.
        :param msg: Incoming WSS Message
        :type msg: dict
        :return:
        """
        send_wss_event(self._q_msg_tx, msg['id'], 'lid_image:starting')
        self._lid_image(msg)
        send_wss_event(self._q_msg_tx, msg['id'], 'lid_image:completed')

    def _lid_image(self, msg: dict) -> None:
        """
        Child class handler for lid image requests.
        To be implemented by the child class.
        This method should capture the image from the lid camera, and upload the resulting image to the Web API.
        :param msg: Incoming WSS Message
        :type msg: dict
        :return:
        """
        raise NotImplementedError

    def lidar_image(self, msg: dict) -> None:
        """
        Process lidar image request.
        Sends related start and finish messages, and calls child class handler to capture
        head images with and without the distance measuring laser enabled.
        :param msg: Incoming WSS Message
        :type msg: dict
        :return:
        """
        send_wss_event(self._q_msg_tx, msg['id'], 'lidar_image:starting')
        self._head_image(msg, msg['settings'][0])
        send_wss_event(self._q_msg_tx, msg['id'], 'lidar_image:completed')

    def motion(self, msg: dict) -> None:
        """
        Process motion request.
        Sends related start and finish messages, and calls child class handler.
        :param msg: Incoming WSS Message
        :type msg: dict
        :return:
        """
        send_wss_event(self._q_msg_tx, msg['id'], msg['action_type'] + ':starting')
        self._motion(msg)
        if self._running_action_cancelled:
            self._send_cancelled_message(msg['id'], msg['action_type'])
        else:
            send_wss_event(self._q_msg_tx, msg['id'], msg['action_type'] + ':completed')

    def _motion(self, msg: dict) -> None:
        """
        Child class handler for motion requests.
        To be implemented by the child class.
        This method should download the specified motion file and execute it.
        :param msg: Incoming WSS Message
        :type msg: dict
        :return:
        """
        raise NotImplementedError

    def run_capture(self, msg: dict) -> None:
        """
        Process capture request.
        Fires up CaptureThread thread to handle image request.
        Interface for GFUI Service
        :param msg: Incoming WSS Message
        :type msg: dict
        :return:
        """
        if self._ok_to_run_action(msg['id'], msg['action_type'], msg['status']):
            if not self._action_thread.is_alive():
                self._action_thread = _ActionThread(self, msg)
                self._action_thread.start()

    def run_puls(self, msg: dict) -> None:
        """
        Process pulse file.
        Fires up MotionThread thread to handle puls file request.
        Interface for GFUI Service
        :param msg: Incoming WSS Message
        :type msg: dict
        :return:
        """
        if self._ok_to_run_action(msg['id'], msg['action_type'], msg['status']):
            if not self._action_thread.is_alive():
                self._action_thread = _ActionThread(self, msg)
                self._action_thread.start()

    def run_settings_report(self, msg: dict) -> None:
        """
        Send settings report.
        Interface for GFUI Service
        :param msg: Incoming WSS Message
        :type msg: dict
        :return:
        """
        if self._ok_to_run_action(msg['id'], msg['action_type'], msg['status']):
            send_report(self._q_msg_tx, msg)
        self.running_action_id = None

    def _shutdown(self) -> None:
        """
        Child class handler for shutting down the machine
        To be implemented by the child class.
        :return:
        """
        raise NotImplementedError

    def _send_cancelled_message(self, action_id: int, action_type: str):
        send_wss_event(self._q_msg_tx, action_id, '%s:cancelled' % action_type)

    def start(self, session: Session, msq_tx_q: Queue) -> None:
        """
        Initialize Machine and Start message handling threads.
        :param session: WSS Session
        :type session: Session
        :param msq_tx_q: WSS Msg Tx Queue
        :type msq_tx_q: Queue
        :return:
        """
        logger.debug('starting')
        self._session = session
        self._q_msg_tx = msq_tx_q
        self._initialize()
        logger.debug('started')

    def stop(self) -> None:
        """
        Stop message handling threads.
        :return:
        """
        logger.debug('stopping')
        self._shutdown()
        logger.debug('stopped')


class _ActionThread(Thread):
    """
    ActionThread
    Responds to incoming WSS events to run an action
    """
    def __init__(self, machine: BaseMachine, msg: dict):
        """
        Initialize ActionThread Dispatcher
        :param machine: Machine object
        :type machine: BaseMachine
        :param msg: Incoming WSS Message
        :type msg: dict
        """
        self._machine = machine
        self._msg = msg
        Thread.__init__(self)

    def run(self) -> None:
        logger.debug('action thread start')
        if self._msg['action_type'] == 'lid_image':
            self._machine.lid_image(self._msg)
        elif self._msg['action_type'] == 'head_image':
            self._machine.head_image(self._msg)
        elif self._msg['action_type'] == 'lidar_image':
            self._machine.lidar_image(self._msg)
        elif self._msg['action_type'] == 'hunt':
            self._machine.hunt(self._msg)
        elif self._msg['action_type'] in ['motion', 'print']:
            self._machine.motion(self._msg)
        self._machine.running_action_id = None
        self._machine.running_action_type = None
        logger.debug('action thread stop')


__all__ = ['BaseMachine']



"""
(C) Copyright 2020
Scott Wiederhold, s.e.wiederhold@gmail.com
https://community.openglow.org

SPDX-License-Identifier:    MIT
"""
import logging
import json
from queue import Queue

from gfutilities._common import *
from gfutilities.configuration import *
from gfutilities.device.basemachine import BaseMachine
from gfutilities.service.authentication import authenticate_machine
from gfutilities.service.websocket import get_session, firmware_check, firmware_download, ws_connect

logger = logging.getLogger(LOGGER_NAME)


class GFUIService:
    """
    Glowforge UI Service Connector
    This connects to the servers specified in the configuration, and interfaces with the service.
    """
    def __init__(self, machine: BaseMachine):
        """
        Class Initialization
        Initialized WSS Queues, and configures logging.
        """
        self.session = None
        self.q_msg_rx = Queue()
        self.q_msg_tx = Queue()
        self.q_capture = Queue()
        self._machine = machine
        logger.info('INITIALIZED')

    def connect(self) -> bool:
        """
        Authenticates machine to service, checks firmware, and establishes WSS session.
        :return: Status of connection
        :rtype: bool
        """
        # Grab a persistent web session.  Once authenticated, this session is used for all web API's.
        self.session = get_session()
        # Authenticate machine
        if not authenticate_machine(self.session):
            return False
        # Check for new firmware
        if get_cfg('FACTORY_FIRMWARE.CHECK'):
            fw = firmware_check(self.session)
            if fw:
                firmware_download(self.session, fw)
        # Establish WebSocket Connection
        if not ws_connect(self.q_msg_rx, self.q_msg_tx):
            return False
        return True

    def run(self) -> None:
        """
        Processes messages from WSS service
        :return:
        """
        self._machine.start(self.session, self.q_msg_tx)
        while True:
            try:
                msg = json.loads(self.q_msg_rx.get())
                logger.info('service action request: %s (%s)' % (msg['action_type'], msg['status']))
                result = 'dispatched'
                if msg['action_type'] == 'settings' and msg['status'] == 'ready':
                    self._machine.run_settings_report(msg)
                elif 'image' in msg['action_type']:
                    self._machine.run_capture(msg)
                elif msg['action_type'] in ('hunt', 'motion', 'print'):
                    self._machine.run_puls(msg)
                else:
                    result = 'ignored'
                logger.info('%s (%s) service action %s' % (msg['action_type'], msg['status'], result))
                self.q_msg_rx.task_done()
            except KeyboardInterrupt:
                break
        self._machine.stop()



"""
(C) Copyright 2020
Scott Wiederhold, s.e.wiederhold@gmail.com
https://community.openglow.org

SPDX-License-Identifier:    MIT
"""
from typing import NamedTuple, Union

LOGGER_NAME = 'openglow'


class MachineSetting(NamedTuple):
    type: type
    in_report: bool
    min_value: Union[int, None]
    max_value: Union[int, None]
    default: Union[int, str, float, None]
    idle: callable = None
    run: callable = None
    cool_down: callable = None


__all__ = [
    # Constants
    'LOGGER_NAME',

    # Named Tuples
    'MachineSetting',
]



from setuptools import setup, find_packages

setup(
    name='gfutilities',
    description='Glowforge Utilities',
    author='Scott Wiederhold',
    author_email='s.e.wiederhold@gmail.com',
    url='https://github.com/ScottW514/Glowforge-Utilities',
    version='0.8.4',
    packages=find_packages(),
    license='MIT',
    long_description=open('README.txt').read(),
    keywords='Glowforge OpenGlow',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.7',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Topic :: Software Development :: Embedded Systems',
    ],
)



"""
(C) Copyright 2020
Scott Wiederhold, s.e.wiederhold@gmail.com
https://community.openglow.org

SPDX-License-Identifier:    MIT
"""
import logging
from queue import Queue
import time
from typing import Any

from gfutilities._common import *
from gfutilities.configuration import get_cfg, set_cfg
from gfutilities.service.websocket import send_wss_event

logger = logging.getLogger(LOGGER_NAME)

MACHINE_SETTINGS = {
    # Settings that are communicated between the service and the device.
    # Mostly just sent to the service from the machine, and it seems to make little difference.
    # Default values are included for those that are not otherwise set (see further down).
    # Throwing garbage values doesn't seem to hurt.  Not sure what the point of this exercise is...
    #
    # "ITEM": (<data type>, <report to service>, <min value>, <max value>, <default value>, <idle callable>,
    # <run callable>, <cool_down callable>)
    "AAdc": MachineSetting(int, True, 0, 1023, 204),  # AirAssist
    "AAfc": MachineSetting(int, True, None, None, 536870912),  # AirAssist
    "AAid": MachineSetting(int, True, 0, 1023, 204),  # AirAssist
    "AAil": MachineSetting(int, True, None, None, 1000),  # AirAssist
    "AAin": MachineSetting(int, True, None, None, 0),  # AirAssist
    "AAix": MachineSetting(int, True, None, None, 0),  # AirAssist
    "AAli": MachineSetting(int, True, None, None, 1000),  # AirAssist
    "AArd": MachineSetting(int, True, None, None, 0),  # AirAssist
    "AArl": MachineSetting(int, True, None, None, 1000),  # AirAssist
    "AArn": MachineSetting(int, True, None, None, 0),  # AirAssist
    "AArx": MachineSetting(int, True, None, None, 0),  # AirAssist
    "AAsd": MachineSetting(int, True, None, None, 0),  # AirAssist
    "AAsl": MachineSetting(int, True, None, None, 1000),  # AirAssist
    "AAsn": MachineSetting(int, True, None, None, 0),  # AirAssist
    "AAsx": MachineSetting(int, True, None, None, 0),  # AirAssist
    "AAtv": MachineSetting(int, True, None, None, 3457),  # AirAssist
    "AAwd": MachineSetting(int, True, None, None, 0),  # AirAssist
    "AAwl": MachineSetting(int, True, None, None, 1000),  # AirAssist
    "AAwn": MachineSetting(int, True, None, None, 0),  # AirAssist
    "AAwx": MachineSetting(int, True, None, None, 0),  # AirAssist
    "AScm": MachineSetting(int, True, None, None, 0),  #
    "AShw": MachineSetting(str, True, None, None, "00000000000000000000000000000000000000000000000000000000000"),  #
    "BAai": MachineSetting(int, True, None, None, 0),  # Board Accel
    "BAar": MachineSetting(int, True, None, None, 0),  # Board Accel
    "BAdb": MachineSetting(int, True, None, None, 5),  # Board Accel
    "BAdl": MachineSetting(int, True, None, None, 5),  # Board Accel
    "BAfa": MachineSetting(int, True, None, None, 0),  # Board Accel
    "BAfo": MachineSetting(int, True, None, None, 0),  # Board Accel
    "BAil": MachineSetting(int, True, None, None, 0),  # Board Accel
    "BAli": MachineSetting(int, True, None, None, 0),  # Board Accel
    "BApc": MachineSetting(int, True, None, None, 500),  # Board Accel
    "BApi": MachineSetting(int, True, None, None, 500),  # Board Accel
    "BApr": MachineSetting(int, True, None, None, 200),  # Board Accel
    "BArl": MachineSetting(int, True, None, None, 0),  # Board Accel
    "BAsi": MachineSetting(int, True, None, None, 2),  # Board Accel
    "BAsr": MachineSetting(int, True, None, None, 2),  # Board Accel
    "BAxi": MachineSetting(int, True, None, None, 0),  # Board Accel
    "BAxr": MachineSetting(int, True, None, None, 0),  # Board Accel
    "BAyi": MachineSetting(int, True, None, None, 0),  # Board Accel
    "BAyr": MachineSetting(int, True, None, None, 0),  # Board Accel
    "BAzi": MachineSetting(int, True, None, None, 0),  # Board Accel
    "BAzr": MachineSetting(int, True, None, None, 0),  # Board Accel
    "BDet": MachineSetting(int, True, None, None, 96),  # e t Beam Detect
    "BDie": MachineSetting(int, True, None, None, 0),  # Beam Detect
    "BDlk": MachineSetting(int, True, None, None, 1966),  # Beam Detect: lambda k
    "BDlt": MachineSetting(int, True, None, None, 6553),  # Beam Detect: lambda t
    "BDpe": MachineSetting(int, True, None, None, 0),  # Beam Detect
    "BDtr": MachineSetting(int, True, None, None, 32),  # Beam Detect: theta r
    "BDtt": MachineSetting(int, True, None, None, 40),  # Beam Detect: theta t
    "BLev": MachineSetting(int, True, None, None, 0),  #
    "BLva": MachineSetting(int, True, None, None, 0),  #
    "BLvb": MachineSetting(int, True, None, None, 0),  #
    "BLvc": MachineSetting(int, True, None, None, 0),  #
    "BTcx": MachineSetting(int, True, None, None, 2147483647),  # Board Temp
    "BTfc": MachineSetting(int, True, None, None, 0),  # Board Temp
    "BTfo": MachineSetting(int, True, None, None, 0),  # Board Temp
    "BTin": MachineSetting(int, True, None, None, -2147483648),  # Board Temp
    "BTix": MachineSetting(int, True, None, None, 2147483647),  # Board Temp
    "BTli": MachineSetting(int, True, None, None, 1000),  # Board Temp
    "BTpd": MachineSetting(int, True, None, None, 2500),  # Board Temp
    "BTrn": MachineSetting(int, True, None, None, -2147483648),  # Board Temp
    "BTrx": MachineSetting(int, True, None, None, 2147483647),  # Board Temp
    "BTvl": MachineSetting(int, True, None, None, 752),  # Board Temp
    "BTwn": MachineSetting(int, True, None, None, -2147483648),  # Board Temp
    "BTwx": MachineSetting(int, True, None, None, 2147483647),  # Board Temp
    "BTxb": MachineSetting(int, True, None, None, 0),  # Board Temp
    "CAid": MachineSetting(int, True, None, None, 0),  #
    "CAwb": MachineSetting(int, True, None, None, 0),  #
    "CCbp": MachineSetting(int, True, None, None, 0),  #
    "CCbt": MachineSetting(int, True, None, None, 0),  #
    "CCfl": MachineSetting(int, True, None, None, 0),  #
    "CCml": MachineSetting(int, True, None, None, 0),  #
    "CCpb": MachineSetting(int, True, None, None, 200000),  #
    "CCro": MachineSetting(int, True, None, None, 5000),  #
    "CCst": MachineSetting(int, True, None, None, 2),  #
    "CCup": MachineSetting(int, True, None, None, 0),  #
    "CCxp": MachineSetting(int, True, None, None, 0),  #
    "CCyp": MachineSetting(int, True, None, None, 0),  #
    "CCzp": MachineSetting(int, True, None, None, 0),  #
    "CDcb": MachineSetting(int, True, None, None, 0),  #
    "CDcc": MachineSetting(int, True, None, None, 1),  #
    "CDsb": MachineSetting(int, True, None, None, 0),  #
    "CDtb": MachineSetting(int, True, None, None, 0),  #
    "CFdn": MachineSetting(int, True, None, None, 0),  #
    "CFdt": MachineSetting(int, True, None, None, 0),  #
    "CFdx": MachineSetting(int, True, None, None, 0),  #
    "CFen": MachineSetting(int, True, None, None, 0),  #
    "CFer": MachineSetting(int, True, None, None, 0),  #
    "CFex": MachineSetting(int, True, None, None, 0),  #
    "CFhp": MachineSetting(int, True, None, None, 0),  #
    "CFhs": MachineSetting(int, True, None, None, 0),  #
    "CFhw": MachineSetting(int, True, None, None, 0),  #
    "CFhx": MachineSetting(int, True, None, None, 0),  #
    "CFki": MachineSetting(int, True, None, None, 0),  #
    "CFkp": MachineSetting(int, True, None, None, 0),  #
    "CFli": MachineSetting(int, True, None, None, 0),  #
    "CFpd": MachineSetting(int, True, None, None, 0),  #
    "CFpp": MachineSetting(int, True, None, None, 0),  #
    "CFps": MachineSetting(int, True, None, None, 1),  #
    "CFrh": MachineSetting(int, True, None, None, 0),  #
    "CFsp": MachineSetting(int, True, None, None, 0),  #
    "CMdt": MachineSetting(int, True, None, None, 18364),  #
    "CMet": MachineSetting(int, True, None, None, 18134),  #
    "CMin": MachineSetting(int, True, None, None, 4008),  #
    "CMix": MachineSetting(int, True, None, None, 50010),  #
    "CMli": MachineSetting(int, True, None, None, 100),  #
    "CMpd": MachineSetting(int, True, None, None, 1000),  #
    "CMrn": MachineSetting(int, True, None, None, 1017),  #
    "CMrx": MachineSetting(int, True, None, None, 31040),  #
    "CMts": MachineSetting(int, True, None, None, 0),  #
    "CMup": MachineSetting(int, True, None, None, 100),  #
    "CMut": MachineSetting(int, True, None, None, 18134),  #
    "CMwn": MachineSetting(int, True, None, None, 1017),  #
    "CMwx": MachineSetting(int, True, None, None, 31040),  #
    "CTin": MachineSetting(int, True, None, None, 391),  #
    "CTix": MachineSetting(int, True, None, None, 935),  #
    "CTrn": MachineSetting(int, True, None, None, 591),  #
    "CTrx": MachineSetting(int, True, None, None, 971),  #
    "CTwn": MachineSetting(int, True, None, None, 591),  #
    "CTwx": MachineSetting(int, True, None, None, 971),  #
    "EFdc": MachineSetting(int, True, None, None, 0),  # Exhaust Fan
    "EFfc": MachineSetting(int, True, None, None, 536870912),  # Exhaust Fan
    "EFid": MachineSetting(int, True, 0, 65535, 0),  # Exhaust Fan
    "EFil": MachineSetting(int, True, None, None, 1000),  # Exhaust Fan
    "EFin": MachineSetting(int, True, None, None, 0),  # Exhaust Fan
    "EFix": MachineSetting(int, True, None, None, 0),  # Exhaust Fan
    "EFli": MachineSetting(int, True, None, None, 1000),  # Exhaust Fan
    "EFrd": MachineSetting(int, True, 0, 65535, 0),  # Exhaust Fan
    "EFrl": MachineSetting(int, True, None, None, 1000),  # Exhaust Fan
    "EFrn": MachineSetting(int, True, None, None, 0),  # Exhaust Fan
    "EFrx": MachineSetting(int, True, None, None, 0),  # Exhaust Fan
    "EFsd": MachineSetting(int, True, None, None, 0),  # Exhaust Fan
    "EFsl": MachineSetting(int, True, None, None, 1000),  # Exhaust Fan
    "EFsn": MachineSetting(int, True, None, None, 0),  # Exhaust Fan
    "EFsx": MachineSetting(int, True, None, None, 0),  # Exhaust Fan
    "EFtv": MachineSetting(int, True, None, None, 0),  # Exhaust Fan
    "EFwd": MachineSetting(int, True, None, None, 0),  # Exhaust Fan
    "EFwl": MachineSetting(int, True, None, None, 1000),  # Exhaust Fan
    "EFwn": MachineSetting(int, True, None, None, 0),  # Exhaust Fan
    "EFwx": MachineSetting(int, True, None, None, 0),  # Exhaust Fan
    "FAdc": MachineSetting(int, True, None, None, 0),  #
    "FAfc": MachineSetting(int, True, None, None, 0),  #
    "FAid": MachineSetting(int, True, None, None, 0),  #
    "FAil": MachineSetting(int, True, None, None, 0),  #
    "FAin": MachineSetting(int, True, None, None, 0),  #
    "FAix": MachineSetting(int, True, None, None, 0),  #
    "FAli": MachineSetting(int, True, None, None, 0),  #
    "FArd": MachineSetting(int, True, None, None, 0),  #
    "FArl": MachineSetting(int, True, None, None, 0),  #
    "FArn": MachineSetting(int, True, None, None, 0),  #
    "FArx": MachineSetting(int, True, None, None, 0),  #
    "FAtv": MachineSetting(int, True, None, None, 0),  #
    "FAwd": MachineSetting(int, True, None, None, 0),  #
    "FAwl": MachineSetting(int, True, None, None, 0),  #
    "FAwn": MachineSetting(int, True, None, None, 0),  #
    "FAwx": MachineSetting(int, True, None, None, 0),  #
    "FEdc": MachineSetting(int, True, None, None, 0),  #
    "FEfc": MachineSetting(int, True, None, None, 0),  #
    "FEid": MachineSetting(int, True, None, None, 0),  #
    "FEil": MachineSetting(int, True, None, None, 0),  #
    "FEin": MachineSetting(int, True, None, None, 0),  #
    "FEix": MachineSetting(int, True, None, None, 0),  #
    "FEli": MachineSetting(int, True, None, None, 0),  #
    "FErd": MachineSetting(int, True, None, None, 0),  #
    "FErl": MachineSetting(int, True, None, None, 0),  #
    "FErn": MachineSetting(int, True, None, None, 0),  #
    "FErx": MachineSetting(int, True, None, None, 0),  #
    "FEta": MachineSetting(int, True, None, None, 0),  #
    "FEtb": MachineSetting(int, True, None, None, 0),  #
    "FEtc": MachineSetting(int, True, None, None, 0),  #
    "FEtd": MachineSetting(int, True, None, None, 0),  #
    "FEwd": MachineSetting(int, True, None, None, 0),  #
    "FEwl": MachineSetting(int, True, None, None, 0),  #
    "FEwn": MachineSetting(int, True, None, None, 0),  #
    "FEwx": MachineSetting(int, True, None, None, 0),  #
    "FIdc": MachineSetting(int, True, None, None, 0),  #
    "FIfc": MachineSetting(int, True, None, None, 0),  #
    "FIid": MachineSetting(int, True, None, None, 0),  #
    "FIil": MachineSetting(int, True, None, None, 0),  #
    "FIin": MachineSetting(int, True, None, None, 0),  #
    "FIix": MachineSetting(int, True, None, None, 0),  #
    "FIli": MachineSetting(int, True, None, None, 0),  #
    "FIrd": MachineSetting(int, True, None, None, 0),  #
    "FIrl": MachineSetting(int, True, None, None, 0),  #
    "FIrn": MachineSetting(int, True, None, None, 0),  #
    "FIrx": MachineSetting(int, True, None, None, 0),  #
    "FIta": MachineSetting(int, True, None, None, 0),  #
    "FItb": MachineSetting(int, True, None, None, 0),  #
    "FItc": MachineSetting(int, True, None, None, 0),  #
    "FItd": MachineSetting(int, True, None, None, 0),  #
    "FIwd": MachineSetting(int, True, None, None, 0),  #
    "FIwl": MachineSetting(int, True, None, None, 0),  #
    "FIwn": MachineSetting(int, True, None, None, 0),  #
    "FIwx": MachineSetting(int, True, None, None, 0),  #
    "FSli": MachineSetting(int, True, None, None, 0),  #
    "FSls": MachineSetting(int, True, None, None, 0),  #
    "FSmd": MachineSetting(int, True, None, None, 0),  #
    "FSms": MachineSetting(int, True, None, None, 0),  #
    "FSpv": MachineSetting(int, True, None, None, 0),  #
    "FStv": MachineSetting(int, True, None, None, 0),  #
    "FTcx": MachineSetting(int, True, None, None, 2147483647),  #
    "FTin": MachineSetting(int, True, None, None, -2147483648),  #
    "FTix": MachineSetting(int, True, None, None, 2147483647),  #
    "FTli": MachineSetting(int, True, None, None, 0),  #
    "FTpd": MachineSetting(int, True, None, None, 2500),  #
    "FTrn": MachineSetting(int, True, None, None, -2147483648),  #
    "FTrx": MachineSetting(int, True, None, None, 2147483647),  #
    "FTvl": MachineSetting(int, True, None, None, 0),  #
    "FTwn": MachineSetting(int, True, None, None, -2147483648),  #
    "FTwx": MachineSetting(int, True, None, None, 2147483647),  #
    "FTxb": MachineSetting(int, True, None, None, 0),  #
    "FVer": MachineSetting(int, True, None, None, 0),  #
    "HAai": MachineSetting(int, True, None, None, 0),  # Head Accel
    "HAdr": MachineSetting(int, True, None, None, 5),  # Head Accel
    "HAdb": MachineSetting(int, True, None, None, 5),  # Head Accel
    "HAdl": MachineSetting(int, True, None, None, 5),  # Head Accel
    "HAfa": MachineSetting(int, True, None, None, 0),  # Head Accel
    "HAfo": MachineSetting(int, True, None, None, 0),  # Head Accel
    "HAil": MachineSetting(int, True, None, None, 0),  # Head Accel
    "HAli": MachineSetting(int, True, None, None, 0),  # Head Accel
    "HApc": MachineSetting(int, True, None, None, 500),  # Head Accel
    "HApi": MachineSetting(int, True, None, None, 500),  # Head Accel
    "HApr": MachineSetting(int, True, None, None, 200),  # Head Accel
    "HArl": MachineSetting(int, True, None, None, 0),  # Head Accel
    "HAsi": MachineSetting(int, True, None, None, 2),  # Head Accel
    "HAsr": MachineSetting(int, True, None, None, 2),  # Head Accel
    "HAxi": MachineSetting(int, True, None, None, 0),  # Head Accel
    "HAxr": MachineSetting(int, True, None, None, 0),  # Head Accel
    "HAyi": MachineSetting(int, True, None, None, 0),  # Head Accel
    "HAyr": MachineSetting(int, True, None, None, 0),  # Head Accel
    "HAzi": MachineSetting(int, True, None, None, 0),  # Head Accel
    "HAzr": MachineSetting(int, True, None, None, 0),  # Head Accel
    "HCae": MachineSetting(int, True, None, None, 0),  # Head Cam: Auto Exposure
    "HCag": MachineSetting(int, True, None, None, 0),  # Head Cam: Auto Gain
    "HCaw": MachineSetting(int, True, None, None, 2),  # Head Cam: AWB
    "HCbb": MachineSetting(int, True, None, None, 1400),  # Head Cam: Blue Balance
    "HCex": MachineSetting(int, True, None, None, 3000),  # Head Cam: Exposure
    "HCga": MachineSetting(int, True, None, None, 30),  # Head Cam: Gain
    "HChf": MachineSetting(int, True, 0, 1, 1),  # Head Cam: Horizontal Flip
    "HCil": MachineSetting(int, False, None, None, None),  # Head Cam Illumination
    "HCrb": MachineSetting(int, True, None, None, 1100),  # Head Cam: Red Balance
    "HCvf": MachineSetting(int, True, 0, 1, 0),  # Head Cam: Vertical Flip
    "HEfv": MachineSetting(int, True, None, None, 199234110),  # Head: Firmware Version Head
    "HEhl": MachineSetting(int, True, None, None, 0),  # Head
    "HEid": MachineSetting(int, True, None, None, 1100),  # Head
    "HEsn": MachineSetting(int, True, None, None, 123456789),  # Serial Number Head
    "HEua": MachineSetting(int, True, None, None, 841174597),  # Head
    "HEub": MachineSetting(int, True, None, None, 5472275),  # Head
    "HEuc": MachineSetting(int, True, None, None, 19),  # Head
    "HEul": MachineSetting(int, True, None, None, 0),  # Head
    "HEwl": MachineSetting(int, True, None, None, 0),  # Head
    "HIix": MachineSetting(int, True, None, None, 1023),  #
    "HIpd": MachineSetting(int, True, None, None, 100),  #
    "HIrx": MachineSetting(int, True, None, None, 1023),  #
    "HTcx": MachineSetting(int, True, None, None, 2147483647),  # Head Temp
    "HTfc": MachineSetting(int, True, None, None, 0),  # Head Temp
    "HTfo": MachineSetting(int, True, None, None, 0),  # Head Temp
    "HTin": MachineSetting(int, True, None, None, -2147483648),  # Head Temp
    "HTix": MachineSetting(int, True, None, None, 2147483647),  # Head Temp
    "HTli": MachineSetting(int, True, None, None, 1000),  # Head Temp
    "HTpd": MachineSetting(int, True, None, None, 2500),  # Head Temp
    "HTrn": MachineSetting(int, True, None, None, -2147483648),  # Head Temp
    "HTrx": MachineSetting(int, True, None, None, 2147483647),  # Head Temp
    "HTvl": MachineSetting(int, True, None, None, 768),  # Head Temp
    "HTwn": MachineSetting(int, True, None, None, -2147483648),  # Head Temp
    "HTwx": MachineSetting(int, True, None, None, 2147483647),  # Head Temp
    "HTxb": MachineSetting(int, True, None, None, 0),  # Head Temp
    "HVcr": MachineSetting(int, True, None, None, 0),  #
    "HVli": MachineSetting(int, True, None, None, 0),  #
    "HVvo": MachineSetting(int, True, None, None, 0),  #
    "IAid": MachineSetting(int, True, None, None, 0),  #
    "IFdc": MachineSetting(int, True, None, None, 0),  # Intake Fan
    "IFdm": MachineSetting(int, True, None, None, 0),  # Intake Fan
    "IFfc": MachineSetting(int, True, None, None, 536870912),  # Intake Fan
    "IFid": MachineSetting(int, True, 0, 65535, 0),  # Intake Fan: Idle PWM
    "IFil": MachineSetting(int, True, None, None, 1000),  # Intake Fan
    "IFin": MachineSetting(int, True, None, None, 0),  # Intake Fan
    "IFix": MachineSetting(int, True, None, None, 0),  # Intake Fan
    "IFli": MachineSetting(int, True, None, None, 1000),  # Intake Fan
    "IFrd": MachineSetting(int, True, 0, 65535, 0),  # Intake Fan: Run PWM
    "IFrl": MachineSetting(int, True, None, None, 1000),  # Intake Fan
    "IFrn": MachineSetting(int, True, None, None, 0),  # Intake Fan
    "IFrx": MachineSetting(int, True, None, None, 0),  # Intake Fan
    "IFsd": MachineSetting(int, True, None, None, 0),  # Intake Fan
    "IFsl": MachineSetting(int, True, None, None, 1000),  # Intake Fan
    "IFsn": MachineSetting(int, True, None, None, 0),  # Intake Fan
    "IFsx": MachineSetting(int, True, None, None, 0),  # Intake Fan
    "IFta": MachineSetting(int, True, None, None, 40882666),  # Intake Fan
    "IFtb": MachineSetting(int, True, None, None, 40424667),  # Intake Fan
    "IFwd": MachineSetting(int, True, 0, 65535, 0),  # Intake Fan
    "IFwl": MachineSetting(int, True, None, None, 1000),  # Intake Fan
    "IFwn": MachineSetting(int, True, None, None, 0),  # Intake Fan
    "IFwx": MachineSetting(int, True, None, None, 0),  # Intake Fan
    "IMct": MachineSetting(int, True, None, None, 3500),  #
    "IMfs": MachineSetting(int, True, None, None, 2),  #
    "IMjq": MachineSetting(int, True, None, None, 40),  #
    "IMmd": MachineSetting(int, True, None, None, 509649408),  #
    "IRli": MachineSetting(int, True, None, None, 0),  #
    "IRva": MachineSetting(int, True, None, None, 32),  #
    "IRvb": MachineSetting(int, True, None, None, 32),  #
    "IRvc": MachineSetting(int, True, None, None, 35),  #
    "IRvd": MachineSetting(int, True, None, None, 36),  #
    "ITcx": MachineSetting(int, True, None, None, 2147483647),  # Interconnect Temp
    "ITfc": MachineSetting(int, True, None, None, 0),  # Interconnect Temp
    "ITfo": MachineSetting(int, True, None, None, 0),  # Interconnect Temp
    "ITin": MachineSetting(int, True, None, None, -2147483648),  # Interconnect Temp
    "ITix": MachineSetting(int, True, None, None, 2147483647),  # Interconnect Temp
    "ITli": MachineSetting(int, True, None, None, 1000),  # Interconnect Temp
    "ITpd": MachineSetting(int, True, None, None, 5000),  # Interconnect Temp
    "ITrn": MachineSetting(int, True, None, None, -2147483648),  # Interconnect Temp
    "ITrx": MachineSetting(int, True, None, None, 2147483647),  # Interconnect Temp
    "ITvl": MachineSetting(int, True, None, None, 7904),  # Interconnect Temp
    "ITwn": MachineSetting(int, True, None, None, -2147483648),  # Interconnect Temp
    "ITwx": MachineSetting(int, True, None, None, 2147483647),  # Interconnect Temp
    "ITxb": MachineSetting(int, True, None, None, 0),  # Interconnect Temp
    "LAai": MachineSetting(int, True, None, None, 0),  # Lid Accel
    "LAar": MachineSetting(int, True, None, None, 0),  # Lid Accel
    "LAdb": MachineSetting(int, True, None, None, 5),  # Lid Accel
    "LAdl": MachineSetting(int, True, None, None, 5),  # Lid Accel
    "LAfa": MachineSetting(int, True, None, None, 0),  # Lid Accel
    "LAfo": MachineSetting(int, True, None, None, 0),  # Lid Accel
    "LAil": MachineSetting(int, True, None, None, 0),  # Lid Accel
    "LAli": MachineSetting(int, True, None, None, 0),  # Lid Accel
    "LApc": MachineSetting(int, True, None, None, 500),  # Lid Accel
    "LApi": MachineSetting(int, True, None, None, 500),  # Lid Accel
    "LApr": MachineSetting(int, True, None, None, 200),  # Lid Accel
    "LArl": MachineSetting(int, True, None, None, 0),  # Lid Accel
    "LAsi": MachineSetting(int, True, None, None, 2),  # Lid Accel
    "LAsr": MachineSetting(int, True, None, None, 2),  # Lid Accel
    "LAxi": MachineSetting(int, True, None, None, 0),  # Lid Accel
    "LAxr": MachineSetting(int, True, None, None, 0),  # Lid Accel
    "LAyi": MachineSetting(int, True, None, None, 0),  # Lid Accel
    "LAyr": MachineSetting(int, True, None, None, 0),  # Lid Accel
    "LAzi": MachineSetting(int, True, None, None, 0),  # Lid Accel
    "LAzr": MachineSetting(int, True, None, None, 0),  # Lid Accel
    "LCae": MachineSetting(int, True, None, None, 0),  # Lid Cam: Auto Exposure
    "LCag": MachineSetting(int, True, None, None, 0),  # Lid Cam: Auto Gain
    "LCaw": MachineSetting(int, True, None, None, 2),  # Lid Cam: AWB
    "LCbb": MachineSetting(int, True, None, None, 1400),  # Lid Cam: Blue Balance
    "LCex": MachineSetting(int, True, None, None, 3000),  # Lid Cam: Exposure
    "LCfl": MachineSetting(int, True, None, None, 0),  # Lid Cam: Flash On
    "LCfm": MachineSetting(int, True, None, None, 2),  # Lid Cam: Flash Mode
    "LCga": MachineSetting(int, True, None, None, 30),  # Lid Cam: Gain
    "LChf": MachineSetting(int, True, 0, 1, 1),  # Lid Cam: Horizontal Flip
    "LCrb": MachineSetting(int, True, None, None, 1100),  # Lid Cam: Red Balance
    "LCvf": MachineSetting(int, True, 0, 1, 0),  # Lid Cam: Vertical Flip
    "LLev": MachineSetting(int, True, None, None, 0),  # Lid LED
    "LLsp": MachineSetting(int, True, None, None, 28),  # Speed Lid LED
    "LLvl": MachineSetting(int, True, None, None, 132),  # Brightness Lid LED
    "LTcx": MachineSetting(int, True, None, None, 2147483647),  # Lid Temp
    "LTfc": MachineSetting(int, True, None, None, 0),  # Lid Temp
    "LTfo": MachineSetting(int, True, None, None, 0),  # Lid Temp
    "LTin": MachineSetting(int, True, None, None, -2147483648),  # Lid Temp
    "LTix": MachineSetting(int, True, None, None, 2147483647),  # Lid Temp
    "LTli": MachineSetting(int, True, None, None, 1000),  # Lid Temp
    "LTpd": MachineSetting(int, True, None, None, 5000),  # Lid Temp
    "LTrn": MachineSetting(int, True, None, None, -2147483648),  # Lid Temp
    "LTrx": MachineSetting(int, True, None, None, 2147483647),  # Lid Temp
    "LTvl": MachineSetting(int, True, None, None, 0),  # Lid Temp
    "LTwn": MachineSetting(int, True, None, None, -2147483648),  # Lid Temp
    "LTwx": MachineSetting(int, True, None, None, 2147483647),  # Lid Temp
    "LTxb": MachineSetting(int, True, None, None, 0),  # Lid Temp
    "MCbc": MachineSetting(int, True, None, None, 1),  # BaseMachine Control: Boot attempts
    "MCbf": MachineSetting(int, True, None, None, 1),  # BaseMachine Control: Boot Flags
    "MCbr": MachineSetting(int, True, None, None, 320),  # BaseMachine Control: Manufacturing Info
    "MCbu": MachineSetting(int, True, None, None, 20576),  # BaseMachine Control: Boot Fuses
    "MCcl": MachineSetting(int, True, None, None, 1000),  # BaseMachine Control
    "MCcn": MachineSetting(int, True, None, None, 0),  # BaseMachine Control
    "MCcp": MachineSetting(int, True, None, None, 2500),  # BaseMachine Control
    "MCct": MachineSetting(int, True, None, None, 49156),  # BaseMachine Control
    "MCcx": MachineSetting(int, True, None, None, 65000),  # BaseMachine Control
    "MCdb": MachineSetting(int, True, None, None, 0),  # BaseMachine Control
    "MCds": MachineSetting(int, True, None, None, 0),  # BaseMachine Control
    "MCdt": MachineSetting(int, True, None, None, None),  # BaseMachine Control: BaseMachine Time
    "MCdv": MachineSetting(str, True, None, None, '370ee0b4dad80d10d088b6b65fd10a9144eff5a5'),  # BaseMachine Control: Software Version
    "MCip": MachineSetting(str, True, None, None, '127.0.0.1'),  # BaseMachine Control: IP Address
    "MCla": MachineSetting(float, True, None, None, 0.13574219),  # BaseMachine Control
    "MClb": MachineSetting(float, True, None, None, 0.032226562),  # BaseMachine Control
    "MClc": MachineSetting(float, True, None, None, 0.010742188),  # BaseMachine Control
    "MCld": MachineSetting(int, True, None, None, 0),  # BaseMachine Control
    "MCma": MachineSetting(int, True, None, None, 0),  # BaseMachine Control
    "MCmi": MachineSetting(int, True, None, None, 0),  # BaseMachine Control
    "MCov": MachineSetting(str, True, None, None, '1.12.3-12'),  # BaseMachine Control: Firmware Version
    "MCpt": MachineSetting(int, True, None, None, 4),  # BaseMachine Control
    "MCrc": MachineSetting(int, True, None, None, 2),  # BaseMachine Control
    "MCsn": MachineSetting(int, True, None, None, None),  # BaseMachine Control: Serial Number
    "MCsv": MachineSetting(str, True, None, None, "2.3.0"),  # BaseMachine Control: Settings Version
    "MCsw": MachineSetting(int, True, None, None, 27),  # BaseMachine Control
    "MCtc": MachineSetting(int, True, None, None, 0),  # BaseMachine Control
    "MCut": MachineSetting(float, True, None, None, 0.26666668),  # BaseMachine Control
    "NIcc": MachineSetting(int, True, None, None, 1),  # Network Interface
    "NIcf": MachineSetting(int, True, None, None, 0),  # Network Interface
    "NIdc": MachineSetting(int, True, None, None, 0),  # Network Interface
    "NIdf": MachineSetting(int, True, None, None, 0),  # Network Interface
    "NIft": MachineSetting(int, True, None, None, 0),  # Network Interface
    "NIim": MachineSetting(int, True, None, None, 1500),  # MTU Network Interface
    "NIma": MachineSetting(str, True, None, None, "de:ad:be:ef:00:00"),  # MAC Network Interface
    "NIrb": MachineSetting(float, True, None, None, 0.017756462),  # Network Interface
    "NIrd": MachineSetting(int, True, None, None, 0),  # Network Interface
    "NIre": MachineSetting(int, True, None, None, 0),  # Network Interface
    "NIro": MachineSetting(int, True, None, None, 0),  # Network Interface
    "NIrp": MachineSetting(int, True, None, None, 50),  # Network Interface
    "NIrr": MachineSetting(float, True, None, None, 19.5),  # Network Interface
    "NIsa": MachineSetting(int, True, None, None, -57),  # Network Interface
    "NIss": MachineSetting(int, True, None, None, -42),  # Network Interface
    "NItb": MachineSetting(float, True, None, None, 0.0094909668),  # Network Interface
    "NItd": MachineSetting(int, True, None, None, 0),  # Network Interface
    "NIte": MachineSetting(int, True, None, None, 0),  # Network Interface
    "NItp": MachineSetting(int, True, None, None, 60),  # Network Interface
    "NItr": MachineSetting(float, True, None, None, 72.199997),  # Network Interface
    "NIts": MachineSetting(int, True, None, None, 0),  # Network Interface
    "NIuc": MachineSetting(int, True, None, None, 0),  # Network Interface
    "NIuf": MachineSetting(int, True, None, None, 0),  # Network Interface
    "NRab": MachineSetting(int, True, None, None, 131072),  #
    "NRac": MachineSetting(int, True, None, None, 7),  #
    "NRae": MachineSetting(int, True, None, None, 2),  #
    "NRaf": MachineSetting(int, True, None, None, 1000),  #
    "NRaj": MachineSetting(int, True, None, None, 1000),  #
    "NRib": MachineSetting(int, True, None, None, 131072),  #
    "NRic": MachineSetting(int, True, None, None, 3),  #
    "NRie": MachineSetting(int, True, None, None, 2),  #
    "NRif": MachineSetting(int, True, None, None, 1000),  #
    "NRij": MachineSetting(int, True, None, None, 500),  #
    "NRpb": MachineSetting(int, True, None, None, 131072),  #
    "NRpc": MachineSetting(int, True, None, None, 3),  #
    "NRpe": MachineSetting(int, True, None, None, 2),  #
    "NRpf": MachineSetting(int, True, None, None, 1000),  #
    "NRpj": MachineSetting(int, True, None, None, 500),  #
    "NRwb": MachineSetting(int, True, None, None, 131072),  #
    "NRwc": MachineSetting(int, True, None, None, 120),  #
    "NRwe": MachineSetting(int, True, None, None, 2),  #
    "NRwf": MachineSetting(int, True, None, None, 500),  #
    "NRwj": MachineSetting(int, True, None, None, 250),  #
    "PAcc": MachineSetting(int, True, None, None, 853),  # Purge Air
    "PAcm": MachineSetting(int, True, None, None, 0),  # Purge Air
    "PAli": MachineSetting(int, True, None, None, 1000),  # Purge Air
    "PAon": MachineSetting(int, True, None, None, 1),  # On Purge Air
    "PAos": MachineSetting(int, True, None, None, 0),  # Purge Air
    "PApd": MachineSetting(int, True, None, None, 1000),  # Purge Air
    "PCid": MachineSetting(int, True, None, None, 19795),  # ID PIC
    "PRcp": MachineSetting(float, True, None, None, 2.875),  #
    "PRfc": MachineSetting(int, True, None, None, 0),  #
    "PRlc": MachineSetting(int, True, None, None, 0),  #
    "PRls": MachineSetting(int, True, None, None, 0),  #
    "PRnt": MachineSetting(int, True, None, None, 12),  #
    "PRrs": MachineSetting(int, True, None, None, 11206656),  #
    "PRsc": MachineSetting(int, True, None, None, 1),  #
    "PRvs": MachineSetting(int, True, None, None, 141811712),  #
    "PTli": MachineSetting(int, True, None, None, 1000),  # Power Temp
    "PTmn": MachineSetting(int, True, None, None, 0),  # Power Temp
    "PTmx": MachineSetting(int, True, None, None, 4294967295),  # Power Temp
    "PTpd": MachineSetting(int, True, None, None, 5000),  # Power Temp
    "PTvl": MachineSetting(int, True, None, None, 628),  # Power Temp
    "RAid": MachineSetting(int, True, None, None, 0),  #
    "SAid": MachineSetting(int, True, None, None, 0),  # Request ID the settings report is responding to
    "SLto": MachineSetting(int, True, None, None, 300000),  #
    "STfr": MachineSetting(int, False, 1000, 200000, 10000),  # Step frequency
    "TCon": MachineSetting(int, True, None, None, 0),  #
    "TCth": MachineSetting(int, True, None, None, 2147483647),  #
    "TCtt": MachineSetting(int, True, None, None, 0),  #
    "UAid": MachineSetting(int, True, None, None, 0),  #
    "ULst": MachineSetting(int, True, None, None, 0),  #
    "WPon": MachineSetting(int, True, 0, 1, 1),  # On Water Pump
    "WSst": MachineSetting(int, True, None, None, 0),  #
    "WTaf": MachineSetting(int, True, None, None, 3221225472),  # Water Temp
    "WTbf": MachineSetting(int, True, None, None, 3221225472),  # Water Temp
    "WTua": MachineSetting(int, True, None, None, 755),  # Water Temp
    "WTub": MachineSetting(int, True, None, None, 754),  # Water Temp
    "WTva": MachineSetting(int, True, None, None, 754),  # Water Temp
    "WTvb": MachineSetting(int, True, None, None, 751),  # Water Temp
    "XScc": MachineSetting(int, True, None, None, 33),  # Current X Step
    "XSdm": MachineSetting(int, True, None, None, 1),  # Decay Mode X Step
    "XShc": MachineSetting(int, True, None, None, 33),  # Current X Step
    "XSmm": MachineSetting(int, True, None, None, 1),  # Microstepping Mode X Step
    "XSrc": MachineSetting(int, True, None, None, 0),  # Current X Step
    "YScc": MachineSetting(int, True, None, None, 5),  # Current Y Step
    "YSdm": MachineSetting(int, True, None, None, 1),  # Decay Mode Y Step
    "YShc": MachineSetting(int, True, None, None, 5),  # Current Y Step
    "YSmm": MachineSetting(int, True, None, None, 1),  # Microstepping Mode Y Step
    "YSrc": MachineSetting(int, True, None, None, 0),  # Current Y Step
    "ZHin": MachineSetting(int, True, None, None, 180),  # Homing Step Interval MachineSetting(ms) Z Step
    "ZScc": MachineSetting(int, True, None, None, 1),  # Current Z Step
    "ZSen": MachineSetting(int, True, 0, 1, 0),  # Enable Z Step
    "ZShc": MachineSetting(int, True, None, None, 1),  # Current Z Step
    "ZSmd": MachineSetting(int, True, None, None, 1),  # Microstepping Mode Z Step
    "ZSrc": MachineSetting(int, True, None, None, 0),  # Current Z Step
}

_CONFIG_PROVIDERS = {
    # These configuration item values are pulled from the configuration file.
    "MCdv": "FACTORY_FIRMWARE.APP_VERSION",
    "MCov": "FACTORY_FIRMWARE.FW_VERSION",
    "MCsn": "MACHINE.SERIAL",
    "HEfv": "MACHINE.HEAD_FIRMWARE",
    'HEid': "MACHINE.HEAD_ID",
    "HEsn": "MACHINE.HEAD_SERIAL",
}

_REGISTERED_PROVIDERS = {
    # These configuration item values are the result of running the specified function
    "MCdt": time.time,
}


def get_machine_setting(parameter: str) -> Any:
    """
    Return value for machine setting
    :param parameter: Setting
    :type parameter: str
    :return: Value or None if doesn't exist
    """
    setting = MACHINE_SETTINGS.get(parameter)
    if setting is None:
        return None
    return setting.default


def send_report(q: Queue, msg: dict) -> None:
    """
    Send device settings report in the required format.
    :param q: {WSS message queue
    :type q: Queue
    :param msg: WSS Message
    :type msg: dict
    :return:
    """
    settings = ''
    logger.info("START")
    for item, setting in MACHINE_SETTINGS.items():
        value = '' if setting.default is None else str(setting.default)
        if setting.in_report:
            if item == 'SAid':
                value = msg['id']
            if item in _CONFIG_PROVIDERS:
                config_value = get_cfg(_CONFIG_PROVIDERS.get(item))
                if config_value is not None:
                    value = str(config_value)
            if item in _REGISTERED_PROVIDERS:
                value = str(_REGISTERED_PROVIDERS[item]())
            if setting.type is int:
                value = int(float(value))
            elif setting.type is str:
                value = '"{}"'.format(value)
            else:
                value = setting.type(value)
            settings += '"{}":{},'.format(item, value)

    # The service appears to bypass the homing process if we send it an empty settings value.
    # It is assumed that this prevents the machine from re-homing every time it loses and regains
    # the connection to the Glowforge service (a seemingly frequent occurrence).
    if not get_cfg('SETTINGS.SET') and not get_cfg('EMULATOR.BYPASS_HOMING'):
        send_wss_event(q, msg['id'], 'settings:completed',
                       data={'key': 'settings', 'value': '{"values":{%s}}' % settings[:-1]})
        set_cfg('SETTINGS.SET', True)
    else:
        send_wss_event(q, msg['id'], 'settings:completed', data={'key': 'settings', 'value': '{}'})
    logger.info("COMPLETE")


def update_settings(settings: dict) -> None:
    for key, val in settings.items():
        MACHINE_SETTINGS[key] = MACHINE_SETTINGS[key]._replace(**val)
